from django.test import RequestFactory

PROTECTED_VARIABLES = [
    'consumer',
    'element',
    'selectors',
    'session',
    'url',
]


class Reflex:
    def __init__(self, consumer, url, element, selectors):
        self.consumer = consumer
        self.url = url
        self.element = element
        self.selectors = selectors
        self.session = consumer.scope['session']

    def get_channel_id(self):
        '''
        Override this to make the reflex send to a different channel
        other than the session_key of the user
        '''
        return self.session.session_key

    @property
    def request(self):
        factory = RequestFactory()
        request = factory.get(self.url)
        request.session = self.consumer.scope['session']
        request.user = self.consumer.scope['user']
        return request



from django import template
from django.template import Template

register = template.Library()


@register.tag(name="raw")
def raw(parser, token):
    nodelist = parser.parse(('endraw',))
    parser.delete_first_token()
    return RawNode(nodelist)


class RawNode(template.Node):
    def __init__(self, nodelist):
        self.nodelist = nodelist

    def render(self, context):
        output = ''
        for node in self.nodelist:
            if node.token.token_type.name == 'BLOCK':
                template = Template(node.token.contents)
                rendered = template.render(context)
                raw = '{% ' + rendered + ' %}'
            elif node.token.token_type.name == 'VAR':
                raw = '{{ ' + node.token.contents + ' }}'
            else:
                msg ='{} is not yet handled'.format(node.token.token_type.name)
                raise Exception(msg)
            output = output + raw
        return output



from pathlib import Path

from django.apps import apps
from django.core.management.base import BaseCommand, CommandError
from django.template.loader import get_template

from ._base import BaseGenerateCommand


TEMPLATES = {
    '_reflex.py': 'sockpuppet/scaffolds/reflex.py',
    '_controller.js': 'sockpuppet/scaffolds/controller.js',
    '.js': 'sockpuppet/scaffolds/application.js',
    '.py': 'sockpuppet/scaffolds/view.py',
    '.html': 'sockpuppet/scaffolds/template.html'
}


class Command(BaseGenerateCommand):
    help = "Scaffold for reflex. Includes javascript and python."

    def add_arguments(self, parser):
        parser.add_argument(
            'app_name', nargs=1, type=str,
            help='The app where the generated files should be placed'
        )
        parser.add_argument(
            'reflex_name', nargs='?', type=str,
            help='The name of the reflex and javascript controller',
            default='example'
        )

    def handle(self, *args, **options):
        app_name = options['app_name'][0]
        reflex_name = options['reflex_name']

        module_path = self.lookup_app_path(app_name)
        self.module_path = Path(module_path)

        paths = [
            ('reflexes', '_reflex.py'),
            ('javascript', '.js'),
            ('javascript/controllers', '_controller.js'),
            ('views', '.py'),
            ('templates', '.html')
        ]

        for path, suffix in paths:
            template_name = TEMPLATES[suffix]
            template = get_template(template_name)
            rendered = template.render({'reflex_name': reflex_name})
            self.create_file(path, '{}{}'.format(reflex_name, suffix), rendered)

        self.create_file('views', '__init__.py', '')
        self.create_file('reflexes', '__init__.py', '')
        self.call_stdout('Scaffolding generated!', _type='SUCCESS')
        if (self.module_path / 'views.py').exists():
            msg = 'We created a views directory which means that you need to move your initial views there'
            self.call_stdout('')
            self.call_stdout(msg, _type='WARNING')

        self.call_stdout("Last step is to add the view to urls.py", _type='SUCCESS')




class Element:
    def __init__(self, attrs):
        self.attributes = attrs

    @property
    def dataset(self):
        def strip_data(key):
            return key.split('-')[1]

        _dataset = {
            strip_data(key): value for key, value in self.attributes.items()
            if key.startswith('data-')
        }
        return _dataset



from django.views.generic.base import TemplateView


class {{ reflex_name|title }}View(TemplateView):
    template_name = '{{ reflex_name }}.html'

    def get_context_data(self, *args, **kwargs):
        context = super().get_context_data(*args, **kwargs)
        context['count'] = 0
        return context



import re


def camelize(word):
    word = re.sub(
        r'[\s_](.)',
        lambda m: m.group(1).title(),
        word, flags=re.DOTALL
    )
    return word


def camelize_value(value):
    if isinstance(value, list):
        value = [camelize_value(val) for val in value]
    elif isinstance(value, dict):
        value = {camelize(key): camelize_value(val) for key, val in value.items()}
    return value


def classify(word):
    tail = camelize(word[1:])
    head = word[:1].title()
    return '{}{}'.format(head, tail)



from asgiref.sync import async_to_sync
from channels.layers import get_channel_layer
from .utils import camelize, camelize_value


class Channel:
    '''
    Accepts a name which should either be the name of the group or the channel_name
    to which the content will be broadcast to.
    '''

    def __init__(self, name, identifier=''):
        self.identifier = identifier
        self.name = name
        self.operations = self.stub()

    def clear(self):
        self.operations = self.stub()

    def add_operation(self, key, options):
        self.operations[key].append(options)

    def stub(self):
        return {
            'dispatch_event': [],
            'morph': [],
            'inner_html': [],
            'outer_html': [],
            'text_content': [],
            'insert_adjacent_html': [],
            'insert_adjacent_text': [],
            'remove': [],
            'set_value': [],
            'set_attribute': [],
            'remove_attribute': [],
            'add_css_class': [],
            'remove_css_class': [],
            'set_dataset_property': [],
            'set_style': []
        }

    def broadcast(self):
        operations = {
            camelize(key): camelize_value(value)
            for key, value in self.operations.items() if value
        }
        channel_layer = get_channel_layer()
        group_send = async_to_sync(channel_layer.group_send)
        group_send(
            self.name,
            {
                "identifier": self.identifier,
                "type": "message",
                "cableReady": True,
                "operations": operations
            }
        )
        self.clear()

    def dispatch_event(self, options={}):
        '''
        dispatch_event: [{
                name:     "string",
                detail:   "object",
                selector: "string",
            }, ...
        ],
        '''
        self.add_operation('dispatch_event', options)

    def morph(self, options={}):
        '''
        morph: [{
                selector:      "string",
                html:          "string"
                children_only:  true|false,
                permanent_attribute_name: "string",
                focus_selector: "string",
            }, ...
        ],
        '''
        self.add_operation('morph', options)

    def inner_html(self, options={}):
        '''
           inner_html: [{
             selector:      "string",
             focus_selector: "string",
             html:          "string"
           }, ...],
        '''
        self.add_operation('inner_html', options)

    def outer_html(self, options={}):
        '''
           outer_html: [{
             selector:      "string",
             focus_selector: "string",
             html:          "string"
           }, ...],
        '''
        self.add_operation('outer_html', options)

    def text_content(self, options={}):
        '''
           text_content: [{
             selector: "string",
             text:     "string"
           }, ...]
        '''
        self.add_operation('text_content', options)

    def insert_adjacent_html(self, options={}):
        '''
        insert_adjacent_html: [{
            selector:      "string",
            focus_selector: "string",
            position:      "string",
            html:          "string"
        }, ...],
        '''
        self.add_operation('insert_adjacent_html', options)

    def remove(self, options={}):
        '''
        remove: [{
            selector:      "string",
            focus_selector: "string,
        }, ...],
        '''
        self.add_operation('remove', options)

    def remove_attribute(self, options={}):
        '''
        remove_attribute: [{
            selector: "string",
            name:     "string"
        }, ...],
        '''
        self.add_operation('remove_attribute', options)

    def set_attribute(self, options={}):
        '''
        set_attribute: [{
            selector: "string",
            name:     "string",
            value:    "string"
        }, ...],
        '''
        self.add_operation('set_attribute', options)

    def set_value(self, options={}):
        '''
        set_value: [{
            selector: "string",
            value:    "string"
        }, ...],
        '''
        self.add_operation('set_value', options)

    def add_css_class(self, options={}):
        '''
        add_css_class: [{
            selector: "string",
            name:     "string"
        }, ...],
        '''
        self.add_operation('add_css_class', options)

    def remove_css_class(self, options={}):
        '''
        remove_css_class: [{
            selector: "string",
            name:     "string"
        }, ...],
        '''
        self.add_operation('remove_css_class', options)

    def set_dataset_property(self, options):
        '''
        set_dataset_property: [{
            selector: "string",
            name:     "string",
            value:    "string"
        }, ...],
        '''
        self.add_operation('set_dataset_property', options)

    def set_style(self, options):
        '''
        set_style: [{
          selector: "string",
          name:     "string",
          value:    "string"
        }, ...],
        '''
        self.add_operation('set_style', options)



from pathlib import Path

from django.apps import apps
from django.core.management.base import BaseCommand, CommandError


class BaseGenerateCommand(BaseCommand):

    def lookup_app_path(self, app_name):
        try:
            config = apps.get_app_config(app_name)
        except LookupError as e:
            raise CommandError(str(e))

        module_path = config.module.__path__[0]
        return module_path

    def call_stdout(self, msg, _type='WARNING'):
        style = getattr(self.style, _type)
        self.stdout.write(style(msg))

    def create_file(self, folder, filename, contents):
        if folder == '':
            base_path = Path.cwd()
        else:
            base_path = self.module_path

        filepath = base_path / folder / filename
        if filepath.exists():
            partial_path = '/'.join(filepath.parts[-4:])
            self.call_stdout('{} already exists, so it will be skipped'.format(partial_path))
            return

        try:
            filepath.parent.mkdir(parents=True)
        except FileExistsError:
            pass

        with filepath.open(mode='w') as f:
            f.write(contents)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
import os
import re

try:
    from setuptools import setup
except ImportError:
    from distutils.core import setup


def get_version(*file_paths):
    """Retrieves the version from sockpuppet/__init__.py"""
    filename = os.path.join(os.path.dirname(__file__), *file_paths)
    version_file = open(filename).read()
    version_match = re.search(r"^__version__ = ['\"]([^'\"]*)['\"]",
                              version_file, re.M)
    if version_match:
        return version_match.group(1)
    raise RuntimeError('Unable to find version string.')


version = get_version("sockpuppet", "__init__.py")
readme = open('README.md').read()
requirements = open('requirements.txt').readlines()

setup(
    name='django-sockpuppet',
    version=version,
    description="""Helping you use websockets in an effective way in views""",
    long_description=readme,
    author='Jonathan Sundqvist',
    author_email='jonathan@argpar.se',
    url='https://github.com/jonathan-s/django-sockpuppet',
    packages=[
        'sockpuppet',
    ],
    include_package_data=True,
    install_requires=requirements,
    long_description_content_type="text/markdown",
    license="MIT",
    zip_safe=False,
    keywords='django-sockpuppet',
    classifiers=[
        'Development Status :: 4 - Beta',
        'Framework :: Django :: 2.2',
        'Framework :: Django :: 3.0',
        'Intended Audience :: Developers',
        'Environment :: Web Environment',
        'License :: OSI Approved :: BSD License',
        'Natural Language :: English',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
    ],
)



from django.urls import re_path
from channels.auth import AuthMiddlewareStack
from channels.routing import ProtocolTypeRouter, URLRouter
from sockpuppet.consumer import SockpuppetConsumer

socket_patterns = [re_path(r'ws/sockpuppet-sync', SockpuppetConsumer)]

application = ProtocolTypeRouter({
    'websocket': AuthMiddlewareStack(
        URLRouter(socket_patterns)
    ),
})



import time
import json
import logging
from importlib import import_module
from functools import wraps
import inspect
from os import walk, path
import sys
from urllib.parse import urlparse

from asgiref.sync import async_to_sync
from bs4 import BeautifulSoup
from channels.generic.websocket import JsonWebsocketConsumer
from django.apps import apps
from django.urls import resolve
from django.conf import settings

from .channel import Channel
from .reflex import PROTECTED_VARIABLES
from .element import Element
from .utils import classify


logger = logging.getLogger('sockpuppet')


class SockpuppetError(Exception):
    pass


def context_decorator(method, extra_context):
    @wraps(method)
    def wrapped(self, *method_args, **method_kwargs):
        method_kwargs.update(extra_context)
        context = method(self, *method_args, **method_kwargs)
        # if context was picked from cache extra context needs to be added again
        context.update(extra_context)
        return context
    return wrapped


class SockpuppetConsumer(JsonWebsocketConsumer):
    reflexes = {}

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.subscriptions = set()

        if not self.reflexes:
            configs = apps.app_configs.values()
            for config in configs:
                self.load_reflexes_from_config(config)

    def connect(self):
        '''
        We use the user session key as a default channel to publish any events
        '''
        super().connect()
        session = self.scope['session']
        has_session_key = session.session_key

        if not has_session_key:
            # normally there is no session key for anonymous users.
            session.save()

        async_to_sync(self.channel_layer.group_add)(
            session.session_key,
            self.channel_name
        )

        if not has_session_key:
            self.group_send(
                self.scope['session'].session_key,
                {
                    'type': 'message',
                    'meta_type': 'cookie',
                    'key': 'sessionid',
                    'value': session.session_key,
                    'max_age': settings.SESSION_COOKIE_AGE
                }
            )

        logger.debug(
            ':: CONNECT: Channel %s session: %s',
            self.channel_name, session.session_key
        )

    def disconnect(self, *args, **kwargs):
        '''
        When we disconnect we unsubscribe from the user session key.
        '''
        session = self.scope['session']
        async_to_sync(self.channel_layer.group_discard)(
            session.session_key,
            self.channel_name
        )
        logger.debug(
            ':: DISCONNECT: Channel %s session: %s',
            self.channel_name, session.session_key
        )
        super().disconnect(*args, **kwargs)

    def message(self, event):
        logger.debug('Sending data: %s', event)
        self.send(json.dumps(event))

    def group_send(self, recipient, message):
        send = async_to_sync(self.channel_layer.group_send)
        send(recipient, message)

    def load_reflexes_from_config(self, config):
        def append_reflex(module):
            # TODO only import classes that are actually that inherits Reflex
            for classname in dir(module):
                if 'reflex' in classname.lower():
                    ReflexClass = getattr(module, classname)
                    self.reflexes[ReflexClass.__name__] = ReflexClass

        modpath = config.module.__path__[0]

        for dirpath, dirnames, filenames in walk(modpath):
            if dirpath == modpath and 'reflexes.py' in filenames:
                # classes in reflexes.py
                import_path = '{}.reflexes'.format(config.name)
                module = import_module(import_path)

                append_reflex(module)
            elif dirpath == path.join(modpath, 'reflexes'):
                # assumes reflexes folder is placed directly in app.
                import_path = '{config_name}.reflexes.{reflex_file}'

                for filename in filenames:
                    name = filename.split('.')[0]
                    full_import_path = import_path.format(
                        config_name=config.name, reflex_file=name
                    )
                    module = import_module(full_import_path)

                    append_reflex(module)

    def reflex_message(self, data, **kwargs):
        logger.debug('Json: %s', data)
        logger.debug('kwargs: %s', kwargs)
        start = time.perf_counter()

        url = data['url']
        selectors = data['selectors'] if data['selectors'] else ['body']
        target = data['target']
        reflex_name, method_name = target.split('#')
        reflex_name = classify(reflex_name)
        arguments = data['args'] if data.get('args') else []
        element = Element(data['attrs'])
        try:
            ReflexClass = self.reflexes.get(reflex_name)
            reflex = ReflexClass(self, url=url, element=element, selectors=selectors)
            self.delegate_call_to_reflex(reflex, method_name, arguments)
        except Exception as e:
            error = '{}: {}'.format(e.__class__.__name__, str(e))
            msg = 'SockpuppetConsumer failed to invoke {target}, with url {url}, {message}'.format(
                target=target, url=url, message=error
            )
            self.broadcast_error(msg, data, reflex)
            _, _, traceback = sys.exc_info()
            exc = SockpuppetError(msg)
            raise exc.with_traceback(traceback)

        try:
            self.render_page_and_broadcast_morph(reflex, selectors, data)
        except Exception as e:
            error = '{}: {}'.format(e.__class__.__name__, str(e))
            message = 'SockpuppetConsumer failed to re-render {url} {message}'.format(
                url=url, message=error
            )
            self.broadcast_error(message, data, reflex)
            _, _, traceback = sys.exc_info()
            exc = SockpuppetError(msg)
            raise exc.with_traceback(traceback)

        logger.debug('Reflex took %6.2fms', (time.perf_counter() - start) * 1000)

    def _get_channelname(self, channel_name):
        try:
            # StimulusReflex sends the channel name in the format
            # of a json blob for name.
            name = json.loads(channel_name)
            name = name['channel'].replace('::', '-')
        except json.decoder.JSONDecodeError:
            name = channel_name
        return name

    def subscribe(self, data, **kwargs):
        name = self._get_channelname(data['channelName'])
        async_to_sync(self.channel_layer.group_add)(
            name,
            self.channel_name
        )

    def unsubscribe(self, data, **kwargs):
        name = self._get_channelname(data['channelName'])
        async_to_sync(self.channel_layer.group_discard)(
            name,
            self.channel_name
        )

    def receive_json(self, data, **kwargs):
        message_type = data.get('type')
        if message_type is None and data.get('target'):
            self.reflex_message(data, **kwargs)
        elif message_type == 'subscribe':
            self.subscribe(data, **kwargs)
        elif message_type == 'unsubscribe':
            self.unsubscribe(data, **kwargs)
        else:
            print('Unsupported')

    def render_page_and_broadcast_morph(self, reflex, selectors, data):
        html = self.render_page(reflex)
        if html:
            self.broadcast_morphs(selectors, data, html, reflex)

    def render_page(self, reflex):
        parsed_url = urlparse(reflex.url)
        url_path = parsed_url.path + parsed_url.query
        resolved = resolve(url_path)
        view = resolved.func

        instance_variables = [
            name for (name, member) in inspect.getmembers(reflex)
            if not name.startswith('__') and name not in PROTECTED_VARIABLES
        ]
        reflex_context = {key: getattr(reflex, key) for key in instance_variables}
        reflex_context['stimulus_reflex'] = True

        original_context_data = view.view_class.get_context_data
        view.view_class.get_context_data = context_decorator(
            view.view_class.get_context_data, reflex_context
        )
        response = view(reflex.request, resolved.args, resolved.kwargs)
        # we've got the response, the function needs to work as normal again
        view.view_class.get_context_data = original_context_data
        reflex.session.save()
        return response.rendered_content

    def broadcast_morphs(self, selectors, data, html, reflex):
        document = BeautifulSoup(html)
        selectors = [selector for selector in selectors if document.select(selector)]

        channel = Channel(reflex.get_channel_id(), identifier=data['identifier'])
        logger.debug('Broadcasting to %s', reflex.get_channel_id())

        for selector in selectors:
            channel.morph({
                'selector': selector,
                'html': ''.join([e.decode_contents() for e in document.select(selector)]),
                'children_only': True,
                'permanent_attribute_name': data['permanent_attribute_name'],
                'stimulus_reflex': {**data}
            })
        channel.broadcast()

    def delegate_call_to_reflex(self, reflex, method_name, arguments):
        method = getattr(reflex, method_name)
        method_signature = inspect.signature(method)
        if len(method_signature.parameters) == 0:
            getattr(reflex, method_name)()
        else:
            getattr(reflex, method_name)(*arguments)

    def broadcast_error(self, message, data, reflex):
        # We may have a sitation where we weren't able to get a reflex
        session_key = reflex.get_channel_id() if reflex else self.scope['session'].session_key
        channel = Channel(session_key)
        data.update({'error': 'message'})
        channel.dispatch_event({
            'name': 'stimulus-reflex:500',
            'detail': {'stimulus_reflex': data}
        })
        channel.broadcast()



import subprocess

from django.template.loader import get_template
from subprocess import CalledProcessError

from ._base import BaseGenerateCommand


class Command(BaseGenerateCommand):
    help = "Generate scaffolding to compile javascript."

    def handle(self, *args, **options):
        install = 'npm install -g add-project-script'
        subprocess.check_call(install, shell=True)
        try:
            build = 'add-project-script -n "build" -v "webpack --mode production"'
            watch = 'add-project-script -n "watch" -v "webpack --watch --info-verbosity verbose"'
            subprocess.check_call(build, shell=True)
            subprocess.check_call(watch, shell=True)
        except CalledProcessError:
            msg = 'Build and watch already in package.json, so skipping these'
            self.call_stdout(msg)

        subprocess.check_call('npm uninstall -g add-project-script', shell=True)

        npm_pkg = 'npm install -save-dev glob sockpuppet-js stimulus_reflex webpack webpack-cli'
        subprocess.check_call(npm_pkg.split(' '))

        template = get_template('sockpuppet/scaffolds/webpack.html')
        rendered = template.render({})
        self.create_file('', 'webpack.config.js', rendered)
        self.call_stdout('Scaffolding generated!', _type='SUCCESS')
